# Functions
from .functions import Avg
from .functions import Concat
from .functions import ConvertTimezone
from .functions import Count
from .functions import CurrentDate
from .functions import DateAdd
from .functions import DateFormat
from .functions import DateSub
from .functions import Extract
from .functions import GroupConcat
from .functions import Max
from .functions import Min
from .functions import StringAgg
from .functions import Sum
from .functions import Weekday

# Properties
from .properties import Array
from .properties import Boolean
from .properties import Constant
from .properties import Null
from .properties import Property

# Structures
from .structures import Between
from .structures import CastOperator
from .structures import Comparision
from .structures import Distinct
from .structures import Interval
from .structures import Join
from .structures import Limit
from .structures import Operator
from .structures import Order
from .structures import Query

# Utils
from .factory import QueryFactory

__all__ = [
    'Avg',
    'Concat',
    'ConvertTimezone',
    'Count',
    'CurrentDate',
    'DateAdd',
    'DateFormat',
    'DateSub',
    'Extract',
    'GroupConcat',
    'Max',
    'Min',
    'StringAgg',
    'Sum',
    'Weekday',
    'Array',
    'Boolean',
    'Constant',
    'Null',
    'Property',
    'Between',
    'CastOperator',
    'Comparision',
    'Distinct',
    'Interval',
    'Join',
    'Limit',
    'Operator',
    'Order',
    'Query',
    'QueryFactory'
]



from enum import Enum
from typing import List, Union

from duckql.functions.base import BaseFunction

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType
from ..properties.constant import Constant
from ..properties.property import Property
from ..properties.boolean import Boolean
from ..structures.interval import Interval
from .comparision import Comparision


class Operator(BaseType):
    class Operation(Enum):
        AND = 'and'
        OR = 'or'
        DIVISION = 'division'
        MULTIPLICATION = 'multiplication'
        SUBTRACTION = 'subtraction'
        ADDITION = 'addition'

        def __str__(self):
            mapping = {
                self.AND: 'AND',
                self.OR: 'OR',
                self.DIVISION: '/',
                self.MULTIPLICATION: '*',
                self.SUBTRACTION: '-',
                self.ADDITION: '+'
            }

            try:
                return mapping[self]
            except KeyError:
                raise ValueError(f"{self.value} doesnt have valid string representation specified!")

    obj: Literal['structures.Operator'] = 'structures.Operator'
    operation: Operation
    properties: List[Union[BaseFunction, Constant, Property, Boolean, Comparision, Interval]]
    alias: str = None

    def to_sql(self) -> str:
        sql = f' {self.operation} '.join(map(str, self.properties))
        sql = f"({sql})"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



from .join import Join
from .order import Order
from .query import Query
from .limit import Limit
from .distinct import Distinct
from .interval import Interval
from .comparision import Comparision
from .between import Between
from .operator import Operator
from .cast_operator import CastOperator

__all__ = [
    "Join",
    "Order",
    "Query",
    "Limit",
    "Distinct",
    "Interval",
    "Comparision",
    "Between",
    "Operator",
    "CastOperator"
]



from enum import Enum
from typing import Union

from pydantic import validator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType
from .operator import Operator
from .comparision import Comparision


class Join(BaseType):
    class Type(Enum):
        LEFT = 'left'
        RIGHT = 'right'
        INNER = 'inner'
        OUTER = 'outer'
        CROSS = 'cross'
        FULL_OUTER = 'full_outer'
        LEFT_OUTER = 'left_outer'
        RIGHT_OUTER = 'right_outer'
        NATURAL = 'natural'

        def __str__(self):
            mapping = {
                self.LEFT: 'LEFT',
                self.RIGHT: 'RIGHT',
                self.INNER: 'INNER',
                self.OUTER: 'OUTER',
                self.CROSS: 'CROSS',
                self.FULL_OUTER: 'FULL OUTER',
                self.LEFT_OUTER: 'LEFT OUTER',
                self.RIGHT_OUTER: 'RIGHT OUTER',
                self.NATURAL: 'NATURAL'
            }

            try:
                return mapping[self]
            except KeyError:
                raise ValueError(f"{self.value} doesnt have valid string representation specified!")

    obj: Literal['structures.Join'] = 'structures.Join'
    entity: str
    type: Type
    on: Union[Comparision, Operator] = []
    alias: str = None

    @validator('entity', pre=True)
    def escape_entity(cls, v):
        return cls.escape(v)

    def to_sql(self) -> str:
        sql = f"{self.type} JOIN {self.entity}"

        if self.alias is not None:
            sql = f"{sql} {self.alias}"

        if self.type != self.Type.NATURAL:
            sql = f"{sql} ON {self.on}"

        return sql



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from .base import BaseFunction


class Now(BaseFunction):
    obj: Literal['functions.Avg'] = 'functions.Now'
    alias: str = None

    def to_sql(self) -> str:
        sql = "NOW()"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



from enum import Enum

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType


class Interval(BaseType):
    class Unit(Enum):
        MICROSECOND = "microsecond"
        SECOND = "second"
        MINUTE = "minute"
        HOUR = "hour"
        DAY = "day"
        WEEK = "week"
        MONTH = "month"
        QUARTER = "quarter"
        YEAR = "year"
        SECOND_MICROSECOND = "second_microsecond"
        MINUTE_MICROSECOND = "minute_microsecond"
        MINUTE_SECOND = "minute_second"
        HOUR_MICROSECOND = "hour_microsecond"
        HOUR_SECOND = "hour_second"
        HOUR_MINUTE = "hour_minute"
        DAY_MICROSECOND = "day_microsecond"
        DAY_SECOND = "day_second"
        DAY_MINUTE = "day_minute"
        DAY_HOUR = "day_hour"
        YEAR_MONTH = "year_month"

    obj: Literal['structures.Interval'] = 'structures.Interval'
    value: int
    unit: Unit

    def to_sql(self) -> str:
        return f"INTERVAL '{self.value} {self.unit.value.upper()}'"



from typing import Union

from pydantic import validator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..structures.distinct import Distinct
from ..functions.base import BaseFunction
from ..properties import Array
from ..properties.property import Property
from ..structures.cast_operator import CastOperator


class StringAgg(BaseFunction):
    obj: Literal['functions.StringAgg'] = 'functions.StringAgg'
    property: Union[Property, BaseFunction, Array, CastOperator, Distinct]
    separator: str
    alias: str = None

    @validator('separator', pre=True)
    def sanitize_separator(cls, v):
        return cls.escape(v)

    def to_sql(self) -> str:
        sql = f"STRING_AGG({self.property}, '{self.separator}')"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



import json
from abc import ABC, abstractmethod
from enum import Enum
from typing import List

from pydantic import BaseModel, root_validator

from duckql.exceptions import ParseError
from duckql.utils import json_parse


class BaseType(BaseModel, ABC):
    class Dialect(Enum):
        POSTGRESQL = 'postgresql'
        MARIADB = 'mariadb'
        MYSQL = 'mysql'

    class Config:
        json_loads = json_parse
        json_dumps = json.dumps

    @root_validator(pre=True)
    def escape_alias(cls, values):
        if 'alias' in values and values['alias'] is not None:
            values['alias'] = cls.escape(values['alias'])
        return values

    @root_validator(pre=True)
    def check_nested_aliases(cls, values):
        if 'properties' in values and str(cls) != "<class 'duckql.structures.query.Query'>":

            if any(map(lambda x: hasattr(x, 'alias') and x.alias is not None, values['properties'])):
                raise ParseError("You can't have alias inside of function!")

        # Absolutely ugly validation, I hate myself for this
        if 'property' in values:
            if isinstance(values['property'], dict):
                if 'alias' in values['property'] and values['property']['alias'] is not None:
                    raise ParseError("You can't have alias inside of function!")
            else:
                if hasattr(values['property'], 'alias') and values['property'].alias is not None:
                    raise ParseError("You can't have alias inside of function!")

        return values

    @abstractmethod
    def to_sql(self) -> str:
        pass

    def __str__(self) -> str:
        return self.to_sql()

    @staticmethod
    def escape(value: str, ignore: List[str] = None) -> str:
        insecure_chars = {
            "\0": "\\0",
            "\r": "\\r",
            "\x08": "\\b",
            "\x09": "\\t",
            "\x1a": "\\z",
            "\n": "\\n",
            "\"": "",
            "'": "",
            "\\": "\\\\",
            ';': '',
        }

        if ignore:
            # Removes ignored chars from insecure_chars dictionary
            # https://stackoverflow.com/a/30351294/4458288
            list(map(insecure_chars.__delitem__, filter(insecure_chars.__contains__, ignore)))

        value = value.translate(
            value.maketrans(insecure_chars)
        )

        to_remove = [
            '--'
        ]

        for item in to_remove:
            value = value.replace(item, '')

        return value



from .boolean import Boolean
from .constant import Constant
from .null import Null
from .property import Property
from .array import Array

__all__ = [
    "Boolean",
    "Constant",
    "Null",
    "Property",
    "Array"
]



from typing import Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from .base import BaseFunction
from ..properties import Constant
from ..properties.property import Property
from ..structures.cast_operator import CastOperator


class InitCap(BaseFunction):
    obj: Literal['functions.InitCap'] = 'functions.InitCap'
    property: Union[Property, BaseFunction, Constant, CastOperator]
    alias: str = None

    def to_sql(self) -> str:
        sql = f"initcap({self.property})"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



from decimal import Decimal
from typing import Union

from pydantic import validator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType


class Constant(BaseType):
    obj: Literal['properties.Constant'] = 'properties.Constant'
    value: Union[str, int, float, Decimal]

    class Config:
        title = "Constant"
        schema_extra = {
            'description': "Object representation of alphanumeric constant"
        }

    @validator('value', pre=True)
    def escape_value(cls, v):
        return cls.escape(str(v))

    def to_sql(self) -> str:
        sql = f"{self.value}"

        if not sql.isnumeric():
            sql = f"'{sql}'"

        return sql



from typing import Union

from duckql.functions.base import BaseFunction

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from duckql.base import BaseType
from duckql.properties.property import Property


class Distinct(BaseType):
    obj: Literal['structures.Distinct'] = 'structures.Distinct'
    property: Union[Property, BaseFunction]

    def to_sql(self) -> str:
        return f"DISTINCT {self.property}"



from typing import Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..functions.base import BaseFunction
from ..properties.constant import Constant
from ..properties.property import Property
from ..structures.cast_operator import CastOperator


class ConvertTimezone(BaseFunction):
    obj: Literal['functions.ConvertTimezone'] = 'functions.ConvertTimezone'
    property: Union[Property, BaseFunction, Constant, CastOperator]
    date_from: Constant
    date_to: Constant
    alias: str = None

    def to_sql(self) -> str:
        """
        TODO: not supported in PostgreSQL
        https://dev.mysql.com/doc/refman/8.0/en/date-and-time-functions.html#function_convert-tz MySQL
        https://mariadb.com/kb/en/library/convert_tz/ MariaDB
        :return: str
        """
        sql = f"CONVERT_TZ({self.property}, {self.date_from}, {self.date_to})"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType


class Null(BaseType):
    obj: Literal['properties.Null'] = 'properties.Null'

    class Config:
        title = "Null"
        schema_extra = {
            'description': "Object representation of NULL"
        }

    def to_sql(self) -> str:
        return 'NULL'



import click

from .structures import Query


@click.group()
def cli():
    pass


@cli.command()
def schema():
    print(Query.schema_json(indent=2))


@cli.command()
@click.argument('filename')
def file(filename):
    my_query = Query.parse_file(filename)
    print(my_query)


if __name__ == '__main__':
    cli()



from typing import Union, List

from duckql.functions.base import BaseFunction
from pydantic import validator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..exceptions import ParseError
from ..base import BaseType
from ..properties.property import Property
from ..properties.constant import Constant


class Between(BaseType):
    obj: Literal['structures.Between'] = 'structures.Between'
    property: Union[BaseFunction, Property, Constant]
    values: List[Union[BaseFunction, Constant, Property]]

    @validator('values', pre=True)
    def check_number_of_attributes(cls, v):
        if len(v) != 2:
            raise ParseError("Between requires exactly two attributes!")
        return v

    def to_sql(self) -> str:
        return f"{self.property} BETWEEN {self.values[0]} AND {self.values[1]}"



from pydantic import validator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType


class Property(BaseType):
    obj: Literal['properties.Property'] = 'properties.Property'
    name: str
    alias: str = None

    class Config:
        title = "Property"
        schema_extra = {
            'description': "Object representation of SQL column/property"
        }

    @validator('name', pre=True)
    def escape_name(cls, v):
        return cls.escape(v)

    def to_sql(self) -> str:
        sql = f"{self.name}"

        if self.alias is not None:
            sql = f'{sql} AS "{self.alias}"'

        return sql



import json
from typing import Dict

from duckql import Query


class QueryFactory:
    @classmethod
    def from_dict(cls, payload: Dict) -> Query:
        return Query.parse_raw(json.dumps(payload))

    @classmethod
    def from_json(cls, payload: str) -> Query:
        return Query.parse_raw(payload)

    @classmethod
    def from_msgpack(cls, payload) -> Query:
        try:
            import msgpack
        except ImportError:
            raise RuntimeError("msgpack is not installed!")

        return cls.from_dict(msgpack.loads(payload))

    @classmethod
    def from_bson(cls, payload) -> Query:
        try:
            import bson
        except ImportError:
            raise RuntimeError("bson is not installed!")

        return cls.from_dict(bson.loads(payload))



from typing import Union

from pydantic import validator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..functions.base import BaseFunction
from ..properties.property import Property
from ..properties.constant import Constant
from ..structures.cast_operator import CastOperator
from ..structures.interval import Interval


class ToChar(BaseFunction):
    obj: Literal['functions.ToChar'] = 'functions.ToChar'
    property: Union[Property, Constant, BaseFunction, CastOperator, Interval]
    format: str
    alias: str = None

    @validator('format', pre=True)
    def sanitize_format(cls, v):
        return cls.escape(v)

    def to_sql(self) -> str:
        sql = f"to_char({self.property}, '{self.format}')"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



import importlib
import json

from duckql.exceptions import ParseError


def custom_parser(data: dict):
    obj = data

    if 'obj' in data:
        bits = data.pop('obj').split('.')

        try:
            module_name = f"duckql.{bits[0]}"
            class_name = bits[1]
        except IndexError:
            raise ParseError("Invalid object!")

        module_ = importlib.import_module(module_name)
        class_ = getattr(module_, class_name)

        obj = class_(**data)

    return obj


def json_parse(s, **kwargs):
    kwargs['object_hook'] = custom_parser
    return json.loads(s, **kwargs)



from typing import Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..functions.base import BaseFunction
from ..properties.constant import Constant
from ..properties.property import Property
from ..structures.cast_operator import CastOperator
from ..structures.interval import Interval


class DateSub(BaseFunction):
    obj: Literal['functions.DateSub'] = 'functions.DateSub'
    property: Union[Constant, Property, BaseFunction, CastOperator]
    interval: Interval
    alias: str = None

    def to_sql(self) -> str:
        sql = f"DATE_SUB({self.property}, {self.interval})"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType


class Limit(BaseType):
    obj: Literal['structures.Limit'] = 'structures.Limit'
    limit: int
    offset: int = None

    def to_sql(self) -> str:
        sql = f"LIMIT {self.limit}"

        if self.offset is not None:
            sql = f"{sql} OFFSET {self.offset}"

        return sql



from enum import Enum
from typing import List

from pydantic import validator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from duckql.exceptions import ParseError
from duckql.base import BaseType
from duckql.properties.array import Array


class Comparision(BaseType):
    class Operation(Enum):
        EQUAL = 'eq'
        NOT_EQUAL = 'neq'
        LOWER = 'lt'
        LOWER_EQUAL = 'lte'
        GREATER = 'gt'
        GREATER_EQUAL = 'gte'
        LIKE = 'like'
        IN = 'in'
        NOT_LIKE = 'nlike'
        NOT_IN = 'nin'
        IS = 'is'
        IS_NOT = 'nis'
        CONTAINS = 'contains'

        def __str__(self):
            mapping = {
                self.EQUAL: '=',
                self.NOT_EQUAL: '!=',
                self.LOWER: '<',
                self.LOWER_EQUAL: '<=',
                self.GREATER: '>',
                self.GREATER_EQUAL: '>=',
                self.LIKE: 'LIKE',
                self.IN: 'IN',
                self.NOT_LIKE: 'NOT LIKE',
                self.NOT_IN: 'NOT IN',
                self.IS: 'IS',
                self.IS_NOT: 'IS NOT',
                self.CONTAINS: 'CONTAINS'
            }

            try:
                return mapping[self]
            except KeyError:
                raise ValueError(f"{self.value} doesnt have valid string representation specified!")

        @classmethod
        def containers(cls) -> List["Comparision.Operation"]:
            return [cls.NOT_IN, cls.IN]

    obj: Literal['structures.Comparision'] = 'structures.Comparision'
    properties: List[BaseType]
    operation: Operation

    @validator('properties', pre=True)
    def check_number_of_properties(cls, v):
        if len(v) != 2:
            raise ParseError("Comparison requires exactly two attributes!")

        return v

    def to_sql(self) -> str:
        sql = f"{self.properties[0]} {self.operation}"

        if not isinstance(self.properties[1], Array) and self.operation in Comparision.Operation.containers():
            sql = f"{sql} ({self.properties[1]})"
        else:
            sql = f"{sql} {self.properties[1]}"

        return f"({sql})"



from typing import List, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..functions.base import BaseFunction
from ..properties.property import Property
from ..properties.constant import Constant
from ..structures.cast_operator import CastOperator


class Concat(BaseFunction):
    obj: Literal['functions.Concat'] = 'functions.Concat'
    properties: List[Union[Property, BaseFunction, Constant, CastOperator]]
    alias: str = None

    def to_sql(self) -> str:
        sql = f'CONCAT({", ".join(map(str, self.properties))})'

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



from .avg import Avg
from .concat import Concat
from .convert_timezone import ConvertTimezone
from .count import Count
from .current_date import CurrentDate
from .date_format import DateFormat
from .date_sub import DateSub
from .extract import Extract
from .group_concat import GroupConcat
from .date_add import DateAdd
from .min_max import Max, Min
from .sum import Sum
from .weekday import Weekday
from .string_agg import StringAgg
from .now import Now
from .to_char import ToChar
from .to_date import ToDate
from .unaccent import Unaccent
from .lower import Lower
from .upper import Upper
from .initcap import InitCap

__all__ = [
    "Avg",
    "Concat",
    "ConvertTimezone",
    "Count",
    "CurrentDate",
    "DateFormat",
    "DateSub",
    "Extract",
    "GroupConcat",
    "DateAdd",
    "Max",
    "Min",
    "Sum",
    "Weekday",
    "StringAgg",
    "Now",
    "ToChar",
    "ToDate",
    "Unaccent",
    "Lower",
    "Upper",
    "InitCap"
]



from typing import List, Union

from .cast_operator import CastOperator
from ..functions.base import BaseFunction
from pydantic import validator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType
from ..properties.property import Property
from .distinct import Distinct
from .operator import Operator
from .comparision import Comparision
from .order import Order
from .limit import Limit
from .join import Join


class Query(BaseType):
    class Config:
        title = 'Query'

    obj: Literal['structures.Query'] = 'structures.Query'
    entity: Union[str, 'Query']
    properties: List[Union[BaseFunction, Property, Distinct, CastOperator]]
    joins: List[Join] = []
    conditions: Union[Operator, Comparision] = None
    having: Union[Operator, Comparision] = None
    order: List[Order] = []
    group: List[Union[Property, BaseFunction]] = []
    limit: Limit = None
    alias: str = None

    @classmethod
    @validator('entity', pre=True)
    def escape_entity(cls, v):
        return cls.escape(v) if isinstance(v, str) else v

    def to_sql(self) -> str:
        sql = f"SELECT {', '.join(map(str, self.properties))} FROM {self.entity}"

        if self.joins:
            sql = f"{sql} {' '.join(map(str, self.joins))}"

        if self.conditions:
            sql = f"{sql} WHERE {self.conditions}"

        if self.group:
            sql = f"{sql} GROUP BY {', '.join(map(str, self.group))}"

        if self.having:
            sql = f"{sql} HAVING {self.having}"

        if self.order:
            sql = f"{sql} ORDER BY {', '.join(map(str, self.order))}"

        if self.limit:
            sql = f"{sql} {self.limit}"

        if self.alias:
            sql = f"({sql}) AS {self.alias}"

        return sql

    def _extract_properties(self, item: BaseType) -> List[Property]:
        if isinstance(item, Property):
            return [item]
        elif isinstance(item, list):
            my_properties = []
            for i in item:
                my_properties.extend(self._extract_properties(i))
            return my_properties
        elif hasattr(item, 'property'):
            return self._extract_properties(getattr(item, 'property', []))
        elif hasattr(item, 'properties'):
            return self._extract_properties(getattr(item, 'properties', []))
        else:
            return []

    @property
    def nested_properties(self) -> List[Property]:
        my_properties = []

        for item in self.properties:
            my_properties.extend(self._extract_properties(item))

        my_properties.extend(self._extract_properties(self.conditions))

        for item in self.order:
            my_properties.extend(self._extract_properties(item))

        for item in self.group:
            my_properties.extend(self._extract_properties(item))

        return my_properties



from enum import Enum
from typing import Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType
from ..properties.property import Property
from ..properties.constant import Constant
from ..functions.base import BaseFunction


class CastOperator(BaseType):
    class DataType(Enum):
        # Boolean
        BOOLEAN = 'boolean'
        BOOL = 'bool'

        # Character
        CHAR = 'char'
        VARCHAR = 'varchar'
        TEXT = 'text'

        # Integers
        SMALLINT = 'smallint'
        INTEGER = 'integer'
        INT = 'int'
        REAL = 'real'

        # Datetime
        TIMESTAMP = 'timestamp'
        DATE = 'date'

    obj: Literal['structures.Distinct'] = 'structures.CastOperator'
    property: Union[Constant, Property, BaseFunction]
    to: DataType
    alias: str = None

    def to_sql(self) -> str:
        sql = f"{self.property}::{self.to.value}"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql



from setuptools import setup


def read_files(files):
    data = []
    for file in files:
        with open(file) as f:
            data.append(f.read())
    return "\n".join(data)


meta = {}
with open('duckql/version.py') as f:
    exec(f.read(), meta)

setup(
    name='duckql',
    version=meta['__version__'],
    packages=[
        'duckql', 'duckql.functions', 'duckql.properties', 'duckql.structures',
    ],
    install_requires=[
        'pydantic==1.*',
        'click==7.*',
        'typing-extensions'
    ],
    url='https://github.com/Sibyx/duckql-python',
    license='MIT',
    author='Jakub Dubec',
    author_email='jakub.dubec@gmail.com',
    description='JSON declarative SQL conversion library',
    long_description=read_files(['README.md', 'CHANGELOG.md']),
    long_description_content_type='text/markdown',
    entry_points={
        "console_scripts": [
            'duckql-cli = duckql.cli:cli'
        ]
    },
    classifiers=[
        # As from https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries',
    ]
)



from enum import Enum
from typing import Union

from duckql.functions.base import BaseFunction

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType
from ..properties.property import Property


class Order(BaseType):
    class Direction(Enum):
        ASC = 'ASC'
        DESC = 'DESC'

    obj: Literal['structures.Order'] = 'structures.Order'
    property: Union[Property, BaseFunction]
    kind: Direction = Direction.ASC

    def to_sql(self) -> str:
        return f"{self.property} {self.kind.value}"



from typing import List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType
from ..properties.constant import Constant


class Array(BaseType):
    obj: Literal['properties.Array'] = 'properties.Array'
    values: List[Constant]

    class Config:
        title = "Array"
        schema_extra = {
            'description': "Object representation of SQL array"
        }

    def to_sql(self) -> str:
        return f'({", ".join(map(str, self.values))})'



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..base import BaseType


class Boolean(BaseType):
    obj: Literal['properties.Boolean'] = 'properties.Boolean'
    value: bool

    class Config:
        title = "Boolean"
        schema_extra = {
            'description': 'Object representation of SQL boolean'
        }

    def to_sql(self) -> str:
        return 'TRUE' if self.value else 'FALSE'



from enum import Enum
from typing import Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..functions.base import BaseFunction
from ..properties.constant import Constant
from ..properties.property import Property
from ..structures.cast_operator import CastOperator
from ..structures.interval import Interval


class Extract(BaseFunction):
    class Unit(Enum):
        CENTURY = 'century'
        DAY = 'day'
        DECADE = 'decade'
        DOW = 'dow'
        DOY = 'doy'
        EPOCH = 'epoch'
        HOUR = 'hour'
        ISODOW = 'isodow'
        ISOYEAR = 'isoyear'
        MICROSECONDS = 'microseconds'
        MILLENNIUM = 'millennium'
        MILLISECONDS = 'milliseconds'
        MINUTE = 'minute'
        MONTH = 'month'
        QUARTER = 'quarter'
        SECOND = 'second'
        TIMEZONE = 'timezone'
        TIMEZONE_HOUR = 'timezone_hour'
        TIMEZONE_MINUTE = 'timezone_minute'
        WEEK = 'week'
        YEAR = 'year'

    obj: Literal['functions.Extract'] = 'functions.Extract'
    property: Union[Constant, Property, BaseFunction, Interval, CastOperator]
    unit: Unit
    alias: str = None

    def to_sql(self) -> str:
        sql = f"EXTRACT({self.unit.value.upper()} FROM {self.property})"

        if self.alias is not None:
            sql = f"{sql} AS {self.alias}"

        return sql
